#include "UsagePredictor.hpp"

#include <cstdlib>

//#define __DEBUG_OUTPUT

UsagePredictor::UsagePredictor (ClusteringAlgorithm *clAlgorithm, DataCollector *dtCollector) {
   clusterAnalyzer = new ClusterAnalyzer(clAlgorithm, dtCollector);
   dataCollector = dtCollector;	
}

// mem in kB (default = 10 mB)
// this call means: can run grid application with
//	- at least rd.cpuUsage free CPU available
//	- at least rd.freeMemory free memory available
bool UsagePredictor::canRunGridApplication(Timestamp timestamp, ResourceData rd, int hours) {
   return canRunGridApplication(timestamp, rd, NULL, hours);
}

// same as above, including the recentDataSample representing the recent usage data (if left null, the
// recent usage data will be obtained via dataCollector)
bool UsagePredictor::canRunGridApplication(Timestamp timestamp, ResourceData rd, UsageData *recentDataSample, int hours) {
   // TODO: this method should be synchronized (Java-like)

   Cluster *predicted;
   double resourceOffset;
   UsageData *ud;
   vector<resource> resources = getAvailableResources();
   vector<resource>::iterator resourcesIterator;

   int startIndex = timestamp.getSecondInDay()/COLLECT_INTERVAL;
   int endIndex;
   int predictedSoFar;
   int totalHoursOfPrediction = (hours * 60 * 60) / COLLECT_INTERVAL;

   UsageData *newRecentDataSample;
   Timestamp original(timestamp);

   clusterAnalyzer->updateClustersIfNeeded(timestamp);

   if (((recentDataSample != NULL) || dataCollector->hasEnoughRecentUsageToPredict(timestamp)) 
         && clusterAnalyzer->hasClusters()) {
#ifdef __DEBUG_OUTPUT
      cout << "LUPA: using clusters to predict usage (" << timestamp.formattedPrint() << ")" << endl;
#endif
 
      for (resourcesIterator = resources.begin(); resourcesIterator != resources.end(); resourcesIterator++) {
         timestamp.setRawTime(original.getRawTime());
         startIndex = timestamp.getSecondInDay()/COLLECT_INTERVAL;
         endIndex = startIndex + totalHoursOfPrediction;
         predictedSoFar = 0;
         newRecentDataSample = recentDataSample;

         if (recentDataSample == NULL) ud = dataCollector->getRecentUsageData();
         else ud = recentDataSample;          

         while (endIndex > SAMPLES_PER_DAY) {
            predicted = clusterAnalyzer->findUsagePattern(*resourcesIterator, startIndex, startIndex + SAMPLES_PER_DAY, newRecentDataSample);
            resourceOffset = ud->resourceAverage(*resourcesIterator, startIndex, startIndex + SAMPLES_PER_DAY);
            if (!predicted->satisfiesResource(*resourcesIterator,
                     rd.getResource(*resourcesIterator),
                     resourceOffset,
                     startIndex + SAMPLES_PER_DAY,
                     2 * SAMPLES_PER_DAY))
               return false;

            // Simulates that the algorithm is now executing in the start of the day of tomorrow
            timestamp = timestamp.addDays(1).beginningOfSameDay();
            newRecentDataSample = new UsageData(timestamp, predicted->getRepresentativeElement());

            // Needed because the Representative Element of the Predicted cluster is in derivative form
            shiftUsageDataByOffset(newRecentDataSample, resourceOffset, *resourcesIterator);
            
            // If it's the first time, copies what we had of real utilization measured
            if (startIndex != 0) {             
               for (int i = 0; i <= startIndex; i++) {
                  newRecentDataSample->getData()[i].setResource(*resourcesIterator, ud->getData()[i].getResource(*resourcesIterator));
               }
            }
            ud = newRecentDataSample;

            // Updating indexes for the next iteration   
            predictedSoFar += SAMPLES_PER_DAY - startIndex;
            startIndex = 0;
            endIndex = startIndex + totalHoursOfPrediction - predictedSoFar;
         }
         
         // The last part of the prediction
         predicted = clusterAnalyzer->findUsagePattern(*resourcesIterator, startIndex, startIndex + SAMPLES_PER_DAY, newRecentDataSample);
         resourceOffset = ud->resourceAverage(*resourcesIterator, startIndex, startIndex + SAMPLES_PER_DAY);
         if (!predicted->satisfiesResource(*resourcesIterator,
                  rd.getResource(*resourcesIterator),
                  resourceOffset,
                  startIndex + SAMPLES_PER_DAY,
                  endIndex + SAMPLES_PER_DAY))
            return false;

      }
   } else {
      //TODO: predict based on the current usage -- predicted usage is the average of the last 4 hours
      if (dataCollector->hasEnoughRecentUsageToPredict(timestamp, getHoursToConsiderInAverage())) {
#ifdef __DEBUG_OUTPUT
         cout << "LUPA: using last " << getHoursToConsiderInAverage() << " hours to predict usage (" << timestamp.formattedPrint() << ")" << endl;
#endif
         ud = dataCollector->getRecentUsageData();
         for (resourcesIterator = resources.begin(); resourcesIterator != resources.end(); resourcesIterator++) {
            resourceOffset = ud->resourceAverage(*resourcesIterator, startIndex + ((24 - getHoursToConsiderInAverage())*(3600/COLLECT_INTERVAL)), startIndex + SAMPLES_PER_DAY);
            if (!resourceSatisfiesMinimum(*resourcesIterator, resourceOffset, rd.getResource(*resourcesIterator)))
               return false;
         }
      } else {
#ifdef __DEBUG_OUTPUT
         cout << "LUPA: unable to predict usage (" << timestamp.formattedPrint() << ")" << endl;
#endif
      }
   }
   return true;
}

// -------- HOW LONG ---------
int UsagePredictor::howLongCanRunGridApplication(Timestamp timestamp, ResourceData rd, UsageData *recentDataSample) {
   int DEFAULT_HOW_LONG = 24;
   int MAX_HOW_LONG = 72;
   int howLong;
   int minimumHowLong = MAX_HOW_LONG;

   Cluster *predicted;
   double resourceOffset;
   UsageData *ud;
   vector<resource> resources = getAvailableResources();
   vector<resource>::iterator resourcesIterator;

   int startIndex = timestamp.getSecondInDay()/COLLECT_INTERVAL;

   UsageData *newRecentDataSample;
   Timestamp original(timestamp);

   clusterAnalyzer->updateClustersIfNeeded(timestamp);

   if (((recentDataSample != NULL) || dataCollector->hasEnoughRecentUsageToPredict(timestamp)) 
         && clusterAnalyzer->hasClusters()) {
#ifdef __DEBUG_OUTPUT
      cout << "LUPA: using clusters to predict usage (" << timestamp.formattedPrint() << ")" << endl;
#endif
      for (resourcesIterator = resources.begin(); resourcesIterator != resources.end(); resourcesIterator++) {
         timestamp.setRawTime(original.getRawTime());
         howLong = 0;
         startIndex = timestamp.getSecondInDay()/COLLECT_INTERVAL;
         newRecentDataSample = recentDataSample;

         if (recentDataSample == NULL) ud = dataCollector->getRecentUsageData();
         else ud = recentDataSample;          

         while (howLong < MAX_HOW_LONG) {
            predicted = clusterAnalyzer->findUsagePattern(*resourcesIterator, startIndex, startIndex + SAMPLES_PER_DAY, newRecentDataSample);
            resourceOffset = ud->resourceAverage(*resourcesIterator, startIndex, startIndex + SAMPLES_PER_DAY);
            int temporaryHowLong = predicted->satisfiesResourceForHowLong(*resourcesIterator,
                  rd.getResource(*resourcesIterator),
                  resourceOffset,
                  startIndex + SAMPLES_PER_DAY,
                  2 * SAMPLES_PER_DAY);
            howLong = howLong + temporaryHowLong;
            
            // If the predicted how long hours is not the maximum possible, we stop the chaining predictions process
            if (temporaryHowLong < ((SAMPLES_PER_DAY - startIndex) * COLLECT_INTERVAL / 3600)) {
               break;
            }
            
            timestamp = timestamp.addDays(1).beginningOfSameDay();
            newRecentDataSample = new UsageData(timestamp, predicted->getRepresentativeElement());

            // Needed because the Representative Element of the Predicted cluster is in derivative form
            shiftUsageDataByOffset(newRecentDataSample, resourceOffset, *resourcesIterator);
            
            // If it's the first time
            if (startIndex != 0) {             
               for (int i = 0; i <= startIndex; i++) {
                  newRecentDataSample->getData()[i].setResource(*resourcesIterator, ud->getData()[i].getResource(*resourcesIterator));
               }
            }
            ud = newRecentDataSample;

            // Updating indexes for the next iteration   
            startIndex = 0;
         }
         if (howLong < minimumHowLong) {
            minimumHowLong = howLong;
         }
      }
      return minimumHowLong;

   } else {
      if (dataCollector->hasEnoughRecentUsageToPredict(timestamp, getHoursToConsiderInAverage())) {
#ifdef __DEBUG_OUTPUT
         cout << "LUPA: using last " << getHoursToConsiderInAverage() << " hours to predict usage (" << timestamp.formattedPrint() << ")" << endl;
#endif
         ud = dataCollector->getRecentUsageData();
         for (resourcesIterator = resources.begin(); resourcesIterator != resources.end(); resourcesIterator++) {
            resourceOffset = ud->resourceAverage(*resourcesIterator, startIndex + ((24 - getHoursToConsiderInAverage())*(3600/COLLECT_INTERVAL)), startIndex + SAMPLES_PER_DAY);
            if (!resourceSatisfiesMinimum(*resourcesIterator, resourceOffset, rd.getResource(*resourcesIterator)))
               return 0;
         }
      } else {
#ifdef __DEBUG_OUTPUT
         cout << "LUPA: unable to predict usage (" << timestamp.formattedPrint() << ")" << endl;
#endif
      }
   }
   return DEFAULT_HOW_LONG;
}

// ------ AVERAGE VALUE ------
double* UsagePredictor::averageResourceUsage(Timestamp timestamp, ResourceData rd, int hours, UsageData *recentDataSample) {
   vector<resource> resources = getAvailableResources();
   vector<resource>::iterator resourcesIterator;

   double *averages = (double *) malloc(resources.size() * sizeof(double));
   int index = 0;
   
   for (resourcesIterator = resources.begin(); resourcesIterator != resources.end(); resourcesIterator++) {
      averages[index] = averageResourceValue(timestamp, *resourcesIterator, rd.getResource(*resourcesIterator), hours, recentDataSample);
      index++;
   }

   return averages;
}

double UsagePredictor::averageResourceValue(Timestamp timestamp, resource r, double minimum, int hours, UsageData *recentDataSample) {
   Cluster *predicted;
   double resourceOffset, totalSum = 0.0;
   int numberOfValues = 0;
   UsageData *ud;

   if (hours <= 0) {
      hours = getHoursToConsiderInAverage();
   }
   
   int startIndex = timestamp.getSecondInDay()/COLLECT_INTERVAL;
   int predictionsRemaining = (hours * 60 * 60) / COLLECT_INTERVAL;
   UsageData *newRecentDataSample = recentDataSample;

   if (recentDataSample == NULL) ud = dataCollector->getRecentUsageData();
   else ud = recentDataSample; 

   clusterAnalyzer->updateClustersIfNeeded(timestamp);

   if (((recentDataSample != NULL) || dataCollector->hasEnoughRecentUsageToPredict(timestamp)) 
         && clusterAnalyzer->hasClusters()) {
#ifdef __DEBUG_OUTPUT
      cout << "LUPA: using clusters to predict usage (" << timestamp.formattedPrint() << ")" << endl;
#endif

      while (predictionsRemaining > 0) {
         predicted = clusterAnalyzer->findUsagePattern(r, startIndex, startIndex + SAMPLES_PER_DAY, newRecentDataSample);
         resourceOffset = ud->resourceAverage(r, startIndex, startIndex + SAMPLES_PER_DAY);

         // Calculates the sum of all measures from startIndex until the end of the day or the end of the
         // hours requested
         for (int index = startIndex; index < SAMPLES_PER_DAY && index < startIndex + predictionsRemaining; index++) {
            resourceOffset += predicted->getRepresentativeElement().getData()[index + SAMPLES_PER_DAY].getResource(r);
            // This value indicates that the machine won't be available
            if (!resourceSatisfiesMinimum(r, resourceOffset, minimum)) {
               return -10.0;
            }
            totalSum += resourceOffset;
            numberOfValues++;
         }

         timestamp = timestamp.addDays(1).beginningOfSameDay();
         newRecentDataSample = new UsageData(timestamp, predicted->getRepresentativeElement());

         shiftUsageDataByOffset(newRecentDataSample, resourceOffset, r);

         // If it's the first time
         if (startIndex != 0) {
            for (int i = 0; i < startIndex; i++) {
               newRecentDataSample->getData()[i].setResource(r, ud->getData()[i + SAMPLES_PER_DAY].getResource(r));
            }
         }

         ud = newRecentDataSample;

         // Updating indexes for the next iteration 
         predictionsRemaining -= SAMPLES_PER_DAY - startIndex;
         startIndex = 0;
      }
      return totalSum / numberOfValues;
   } else {
      //TODO: predict based on the current usage -- predicted usage is the average of the last 4 hours
      if (dataCollector->hasEnoughRecentUsageToPredict(timestamp, getHoursToConsiderInAverage())) {
#ifdef __DEBUG_OUTPUT
         cout << "LUPA: using last " << getHoursToConsiderInAverage() << " hours to predict usage (" << timestamp.formattedPrint() << ")" << endl;
#endif
         ud = dataCollector->getRecentUsageData();
         resourceOffset = ud->resourceAverage(r, startIndex + ((24 - getHoursToConsiderInAverage())*(3600/COLLECT_INTERVAL)), startIndex + SAMPLES_PER_DAY);

         return resourceOffset;
      } else {
#ifdef __DEBUG_OUTPUT
         cout << "LUPA: unable to predict usage (" << timestamp.formattedPrint() << ")" << endl;
#endif
      }
   }
   // This values indicates that LUPA has not enough information to predict
   return -1.0;
}


vector<double> UsagePredictor::getPrediction(Timestamp timestamp, resource r, int hours) {
   return getPrediction(timestamp, r, hours, dataCollector->getRecentUsageData());
}

vector<double> UsagePredictor::getPrediction(Timestamp timestamp, resource r, int hours, UsageData *recentDataSample) {
   Cluster *predicted;
   double resourceOffset;
   UsageData *ud;
   unsigned int retSize = (hours * 3600)/COLLECT_INTERVAL;
   vector<double> ret(retSize);

   int startIndex = timestamp.getSecondInDay()/COLLECT_INTERVAL;

   clusterAnalyzer->updateClustersIfNeeded(timestamp);
   if (((recentDataSample != NULL) || dataCollector->hasEnoughRecentUsageToPredict(timestamp)) && clusterAnalyzer->hasClusters()) {
      cout << "Prevendo no getPrediction()" << endl;
      predicted = clusterAnalyzer->findUsagePattern(r, startIndex, startIndex + SAMPLES_PER_DAY, recentDataSample);
      if (recentDataSample == NULL) ud = dataCollector->getRecentUsageData();
      else ud = recentDataSample;
      resourceOffset = ud->resourceAverage(r, startIndex, startIndex + SAMPLES_PER_DAY);

      /*
       * FIXME: URGENTE !!! essa lógica e a do canRunGridApplication não podem ficar separadas ! O canRun.. tem que usar o getPrediction
       *        pra ver se satisfaz a condição
       */
      for (unsigned int i = 0; i < retSize; i++) {
         resourceOffset += predicted->getRepresentativeElement().getData()[startIndex + SAMPLES_PER_DAY + i].getResource(r);
         ret[i] = resourceOffset;
      }

   } else {
      //TODO: predict based on the current usage
      return ret;
   }
   return ret;

}

void UsagePredictor::shiftUsageDataByOffset(UsageData *usageData, double resourceOffset, resource r) {
   vector<ResourceData>& data = usageData->getData();
   double resourceValue = resourceOffset;

   for (int i = 0; i < 2 * SAMPLES_PER_DAY; i++) {
      resourceValue += data[i].getResource(r);
      data[i].setResource(r, resourceValue);
   }
}
