package grm;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Scanner;

import java.io.File;
import java.io.FileNotFoundException;

import java.lang.Double;

import org.omg.CORBA.ORB;
import org.omg.CosTrading.Property;

import resourceProviders.Lrm;
import resourceProviders.LrmHelper;

/* This algorithms takes an average of the usage for the next hours of CPU and RAM in each node, and the
 * sorts the selectedLrms array in a way that the ones that comes first is the ones that has the 
 * maximum average */

class GreedyAverageResourceUsageSchedulingAlgorithm implements SchedulingAlgorithm {
   
   private ORB orb;

   public GreedyAverageResourceUsageSchedulingAlgorithm(ORB orb) {
      this.orb = orb;
   }

   public void scheduleApplication(ApplicationRequirements applicationRequirements, ArrayList<Property[]> selectedLrms) {
      int index = 0;

      int size = selectedLrms.size();

      double cpuWeight = Scheduler.CPU_WEIGHT;

      double ramWeight = Scheduler.RAM_WEIGHT;

      int hoursOfPrediction = applicationRequirements.getHoursValue();
      double freeCpuNeeded = applicationRequirements.getResourceValue(Resource.CPU_USAGE);
      /* Needed because we want the value for FREE cpu, not USED cpu */
      freeCpuNeeded = 1.0f - freeCpuNeeded;
      double freeRamNeeded = applicationRequirements.getResourceValue(Resource.FREE_RAM);

      ArrayList<Double> weightedAverages;

      ArrayList<Double> freeCpuAverages = new ArrayList<Double>(size);
      ArrayList<Double> freeRamAverages = new ArrayList<Double>(size);

      for (Iterator it = selectedLrms.iterator(); it.hasNext(); ) {
         Property[] lrmProperties = (Property[]) it.next();
         String lrmIor = lrmProperties[0].value.extract_string();

         try {
            Lrm lrm = LrmHelper.narrow(orb.string_to_object(lrmIor));

            double[] averageUsage = lrm.averageResourceUsage(freeCpuNeeded, freeRamNeeded, hoursOfPrediction);
            double averageCpuUsage = averageUsage[0];
            // If the value is -1.0, it means that LUPA could not predict, so we assume the maximum value
            if (averageCpuUsage == -1.0) {
               averageCpuUsage = 0.0;
            }

            double averageFreeRam = averageUsage[1];
            if (averageFreeRam == -1.0) {
               averageFreeRam = Double.MAX_VALUE;
            }

            double averageFreeCpu = (1.0 - averageCpuUsage); // remember that CPU is given in usage percentage

            averageFreeCpu = averageFreeCpu - freeCpuNeeded;
            averageFreeRam = averageFreeRam - freeRamNeeded;

            // If the value is -10.0, it means that, in the prediction, LUPA found some predicted value lesser than
            // the required, so it stoped and signed that the availability will not be enough
            if (averageUsage[0] == -10.0 || averageUsage[1] == -10.0) {
               it.remove();
            } else {
               freeCpuAverages.add(Double.valueOf(averageFreeCpu));
               freeRamAverages.add(Double.valueOf(averageFreeRam));
            }
         } catch (org.omg.CORBA.TRANSIENT transientException) {
            it.remove();
         }
      }

      normalizeValues(freeRamAverages);

      weightedAverages = new ArrayList<Double>(freeCpuAverages.size());

      /* Here, we combine both the freeRam and freeCpu values, so we sort the list
       * with respect to just one value. The weight of each resource in the final
       * value is setted in the configuration file, 0.5 for each by default */
      for (index = 0; index < freeRamAverages.size(); index++) {
         double freeCpu = freeCpuAverages.get(index);
         double freeRam = freeRamAverages.get(index);
         double averageUsage = (cpuWeight * freeCpu) + (ramWeight * freeRam);

         weightedAverages.add(index, Double.valueOf(averageUsage));
      }

      size = selectedLrms.size();

      /* Sorts the selectedLrms array, with the informations in the
       * weightedAverages array (sorted in descending order). 
       * A simple Selection Sort is used */
      Double maxUsage;
      int maxIndex;
      for (index = 0; index < size; index++) {
         maxUsage = Double.valueOf(Double.MIN_VALUE);
         maxIndex = -1;

         for (int j = index; j < size; j++) {
            Double current = weightedAverages.get(j);
            if (current != null && (current.compareTo(maxUsage) > 0)) {
               maxIndex = j;
               maxUsage = current;
            }
         }
         // Shifts the maximum usage to the correct position
         if (maxIndex != -1) {
            Property[] temp = selectedLrms.get(index);
            selectedLrms.set(index, selectedLrms.get(maxIndex));
            selectedLrms.set(maxIndex, temp);
            weightedAverages.set(maxIndex, weightedAverages.get(index));
         }
      }
   }

   /* Normalize the values, so both the cpuUsage and the freeRam values
    * are in the range 0.0-1.0 */
   private void normalizeValues(ArrayList<Double> originalValues) {
      double maxValue = 0;

      for (Iterator it = originalValues.iterator(); it.hasNext(); ) {
         Double value = (Double) it.next();
         if (maxValue < value.doubleValue()) {
            maxValue = value.doubleValue();
         }
      }

      for (int index = 0; index < originalValues.size(); index++ ) {
         Double value = originalValues.get(index);
         double normalized = value.doubleValue() / maxValue;
         originalValues.set(index, Double.valueOf(normalized));
      }
   }

   public String toString() {
      return "GREEDY";
   }
}
