package grm;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.Scanner;

import org.omg.CORBA.ORB;
import org.omg.CosTrading.Property;

import dataTypes.ApplicationExecutionInformation;
import dataTypes.ProcessExecutionInformation;

/* This class represents the Scheduler of InteGrade. Here we choose which scheduling algorithm
 * to use, verify the application requirements, and then submit this requirements for the 
 * appropriate algorithm. The results are then returned do the GRM */

public class Scheduler {

   private ORB orb;

   private static final String configurationFile = "../scheduler.conf";
   
   private SchedulingAlgorithm[] algorithms = new SchedulingAlgorithm[5];

   private SchedulingAlgorithm schedulingAlgorithm;

   private static int ALGORITHM_TYPE = 0;

   public static double CPU_WEIGHT = 0.5;

   public static double RAM_WEIGHT = 0.5;

   public Scheduler(ORB orb) {
      this.orb = orb;

      algorithms[0] = new NoLupaSchedulingAlgorithm(orb);
      algorithms[1] = new CanRunGridApplicationSchedulingAlgorithm(orb);
      algorithms[2] = new HowLongCanRunGridApplicationSchedulingAlgorithm(orb);
      algorithms[3] = new GreedyAverageResourceUsageSchedulingAlgorithm(orb);
      algorithms[4] = new BestFitAverageResourceUsageSchedulingAlgorithm(orb);

   }
   
   /* This method chooses the most appropriates machines to execute an application, based on some heuristic
    *
    * @param applicationExecutionInformation - Informations about the application, like its requirements
    * @param processExecutionInformation - Information about each process of the application
    * @param selectedLrms - The initial list of candidates to execute the application. This list will be modified
    *                       by the scheduler, and then returned to the GRM 
    *
    * @return boolean - Returns true if it managed to find the minimum amount of hosts to execute the application, or
    *                   false in the other case
    *
    * */
   public boolean scheduleApplication(ApplicationExecutionInformation applicationExecutionInformation, 
         ProcessExecutionInformation[] processExecutionInformation, ArrayList<Property[]> selectedLrms ) {

      readConfigurationFile();

      schedulingAlgorithm = algorithms[ALGORITHM_TYPE];

      RequirementsAnalyser analyser = new RequirementsAnalyser(applicationExecutionInformation);

      ApplicationRequirements applicationRequirements = analyser.getApplicationRequirements();

      schedulingAlgorithm.scheduleApplication(applicationRequirements, selectedLrms);

      SchedulingLogger schedulingLogger = new SchedulingLogger(schedulingAlgorithm, applicationRequirements, selectedLrms);

      if (((selectedLrms.size() < processExecutionInformation.length) && applicationExecutionInformation.forceDifferentNodes)
            || (selectedLrms.size() == 0)) {
            
            schedulingLogger.registerLog(SchedulingLogger.UNSUCCESSFUL_SCHEDULING);
            return false;
      }
      schedulingLogger.registerLog(SchedulingLogger.SUCCESSFUL_SCHEDULING);
      return true;
   }

   /* Reads and configuration file to set 3 parameters of the scheduler:
    *   -> The name of the algorithm
    *   -> The weight of the resource CPU_USAGE (only for Greedy and BestFit algorithms)
    *   -> The weight of the resource FREE_RAM (only for Greedy and BestFit algorithms)
    *
    * */
   private void readConfigurationFile() {
      BufferedReader confFile = null;

      try {
         confFile = new BufferedReader((new InputStreamReader(new FileInputStream(configurationFile))));
      } catch (FileNotFoundException e) {
         System.err.println("Scheduler configuration file not found. Going with default configurations");
         return;
      }

      try {
         while (true) {
            String line = null;
            line = confFile.readLine();

            if (line == null) {
               break;
            }

            StringTokenizer st = new StringTokenizer(line);

            if (st.hasMoreTokens()) {
               String option = st.nextToken();

               // Comments
               if (option.startsWith("#")) {
                  continue;
               } else if (option.equals("algorithm_name")) {
                  if (st.hasMoreTokens()) {
                     String algorithmName = st.nextToken();
                     ALGORITHM_TYPE = getAlgorithmType(algorithmName);
                  }
               } else if (option.equals("cpu_weight")) {
                  if (st.hasMoreTokens()) {
                     String cpuWeight = st.nextToken();
                     CPU_WEIGHT = Double.parseDouble(cpuWeight);
                  }
               } else if (option.equals("ram_weight")) {
                  if (st.hasMoreTokens()) {
                     String ramWeight = st.nextToken();
                     RAM_WEIGHT = Double.parseDouble(ramWeight);
                  }
               } 
            }
         }
      } catch (IOException e) {
         System.err.println("A problem occured while reading the scheduler configuration file");
         return;
      }
   }

   private int getAlgorithmType(String algorithmName) {
      if (algorithmName.equals("NO_LUPA")) {
         return 0;
      } else if (algorithmName.equals("CAN_RUN")) {
         return 1;
      } else if (algorithmName.equals("HOW_LONG")) {
         return 2;
      } else if (algorithmName.equals("GREEDY")) {
         return 3;
      } else if (algorithmName.equals("BEST_FIT")) {
         return 4;
      }
      return 0;
   }
}
